# Copyright (c) 2012, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of Intel Corporation nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Python initialization, to run at BITS startup."""

def init():
    # Set up redirection first, before importing anything else, so that any
    # errors in subsequent imports will get captured into the log.
    import redirect
    redirect.redirect()
    import grubcmds
    grubcmds.register()

    import acpi
    import bits
    import os
    try:
        mcfg_resources = acpi.parse_mcfg()
    except acpi.TableParseException as e:
        print "Error occurred initializing PCI Express base from MCFG:"
        print e
    else:
        for mcfg_resource in mcfg_resources:
            if mcfg_resource.Segment == 0:
                if mcfg_resource.HighAddress != 0:
                    print "Error: PCI Express base above 32 bits is unsupported by BITS"
                    break
                bits.pcie_set_base(mcfg_resource.LowAddress)
                os.putenv('pciexbase', '{:#x}'.format(mcfg_resource.LowAddress))
                break
        else:
            print "Error initializing PCI Express base from MCFG: no resource with segment 0"

    import testsuite
    testsuite.init()

    import testacpi
    testacpi.register_tests()
    import testsmrr
    testsmrr.register_tests()

    from cpudetect import cpulib
    cpulib.register_tests()
