/*
Copyright (c) 2010, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

.intel_syntax noprefix
.arch .sse3

#include <grub/symbol.h>
#include "smpequ.h"

.text

#------------------------------------------------------------------------------
FUNCTION (input_u8)
# return U8 value
# parameters -- port:WORD
  push  ebp
  mov   ebp, esp
  push  edx
  mov   dx, [ebp+8]     # port
  in    al, dx
  pop   edx
  leave
  ret

#------------------------------------------------------------------------------
FUNCTION (output_u8)
# return none
# parameters -- port:WORD, val:BYTE
  push  ebp
  mov   ebp, esp
  push  eax
  push  edx
  mov   dx, [ebp+8]     # port
  mov   al, [ebp+0xc]   # val
  out   dx, al
  pop   edx
  pop   eax
  leave
  ret

#------------------------------------------------------------------------------
FUNCTION(rdmsr32)
# return status 0=ok 1=gpf
# parameters -- msr:DWORD, lo_data_addr:PTR DWORD, hi_data_addr:PTR DWORD, status_addr:PTR DWORD

  push  ebp
  mov   ebp, esp
  push  esi
  push  eax
  push  edx
  push  ecx
  mov   ecx, [ebp+8]    # msr
  jmp   try_rdmsr

  .long 0x58475046      # DWORD 'XGPF'
  .long catch_rdmsr     # dword offset catch_rdmsr
try_rdmsr:
  rdmsr
  mov   esi, [ebp+0x14] # status_addr
  mov   [esi], dword ptr 0
  mov   esi, [ebp+0xc]  # lo_data_addr
  mov   [esi], eax
  mov   esi, [ebp+0x10] # hi_data_addr
  mov   [esi], edx
  jmp   rdmsr32_checked_exit

catch_rdmsr:
  mov   esi, [ebp+0x14]  # status_addr
  mov   [esi], DWORD PTR -1

rdmsr32_checked_exit:
  pop   ecx
  pop   edx
  pop   eax
  pop   esi
  leave
  ret

#------------------------------------------------------------------------------
FUNCTION(wrmsr32)
# return status 0=ok 1=gpf
# parameters -- msr:DWORD, lo_data:DWORD, hi_data:DWORD, status_addr:PTR DWORD

  push  ebp
  mov   ebp, esp
  push  esi
  push  eax
  push  edx
  push  ecx
  mov   ecx, [ebp+8]    # msr
  mov   eax, [ebp+0xc]  # lo_data
  mov   edx, [ebp+0x10] # hi_data
  jmp   try_wrmsr

  .long 0x58475046      # DWORD 'XGPF'
  .long catch_wrmsr     # dword offset catch_wrmsr
try_wrmsr:
  wrmsr
  mov   esi, [ebp+0x14] # status_addr
  mov   [esi], dword ptr 0
  jmp   wrmsr32_checked_exit

catch_wrmsr:
  mov   esi, [ebp+0x14]  # status_addr
  mov   [esi], DWORD PTR -1

wrmsr32_checked_exit:
  pop   ecx
  pop   edx
  pop   eax
  pop   esi
  leave
  ret

#------------------------------------------------------------------------------
FUNCTION(cpuid32)
# parameters -- func:DWORD, eax:PTR DWORD, ebx:PTR DWORD, ecx:PTR DWORD, edx:PTR DWORD

    push  ebp
    mov   ebp,esp
    push  eax
    push  ebx
    push  ecx
    push  edx
    push  esi
    mov   eax, [ebp+0x8]
    cpuid
    mov   esi, [ebp+0xc]  # addr eax
    mov   [esi], eax
    mov   esi, [ebp+0x10] # addr ebx
    mov   [esi], ebx
    mov   esi, [ebp+0x14] # addr ecx
    mov   [esi], ecx
    mov   esi, [ebp+0x18] # addr edx
    mov   [esi], edx
    pop   esi
    pop   edx
    pop   ecx
    pop   ebx
    pop   eax
    leave
    ret

#------------------------------------------------------------------------------
FUNCTION(cpuid32_indexed)
# parameters -- func:DWORD, index:DWORD, eax:PTR DWORD, ebx:PTR DWORD, ecx:PTR DWORD, edx:PTR DWORD

    push  ebp
    mov   ebp,esp
    push  eax
    push  ebx
    push  ecx
    push  edx
    push  esi
    mov   eax, [ebp+0x8]
    mov   ecx, [ebp+0xc]
    cpuid
    mov   esi, [ebp+0x10] # addr eax
    mov   [esi], eax
    mov   esi, [ebp+0x14] # addr ebx
    mov   [esi], ebx
    mov   esi, [ebp+0x18] # addr ecx
    mov   [esi], ecx
    mov   esi, [ebp+0x1C] # addr edx
    mov   [esi], edx
    pop   esi
    pop   edx
    pop   ecx
    pop   ebx
    pop   eax
    leave
    ret

#------------------------------------------------------------------------------
FUNCTION(get_idt_base)
#   local stack variable idtr:FWORD
   push ebp
   mov  ebp, esp
   add  esp, dword ptr -8
   sidt [ebp-6]
   mov  eax, [ebp-4]
   leave
   ret

#------------------------------------------------------------------------------
FUNCTION(get_idt_limit)
#   local stack variable idtr:FWORD
   push ebp
   mov  ebp, esp
   add  esp, 0xfffffff8
   sidt [ebp-6]
   mov  ax, [ebp-6]
   leave
   ret

#------------------------------------------------------------------------------
FUNCTION(set_idtr)
  push ebp
  mov ebp, esp
  push eax
  mov eax, [ebp+8]
  lidt [eax]
  pop eax
  leave
  ret

#------------------------------------------------------------------------------
FUNCTION(get_idtr)

  push ebp
  mov ebp, esp
  push eax
  mov eax, [ebp+8]
  sidt [eax]
  pop eax
  leave
  ret

#------------------------------------------------------------------------------
FUNCTION(set_protected_mode_exception_handler)
# paramters gate_number:DWORD, gate_offset:DWORD, gate_selector:WORD

# Initialize IDT descriptor for gate specified
  push  ebp
  mov   ebp, esp
  push  eax
  push  ebx
  push  edi
  mov   ebx, [ebp+8]        # ID-number for the gate
  call  get_idt_base
  lea   edi, [eax][ebx *8]  # gate's offset-address
  mov   ebx, [ebp+0xc]
  mov   [edi][0], bx        # entry-point's loword
  shr   ebx, 16
  mov   [edi][6], bx        # entry-point's hiword
  mov   bx, [ebp+0x10]      # code-segment selector
  mov   [edi][2], bx        # code-segment selector
  movw  [edi][4], 0x8E00    # 386 interrupt-gate
  pop   edi
  pop   ebx
  pop   eax
  leave
  ret

#------------------------------------------------------------------------------
FUNCTION(get_protected_mode_exception_handler)
# gate_number:DWORD, offset_ptr:PTR DWORD, selector_ptr:PTR WORD
# Initialize IDT descriptor for gate specified
  push  ebp
  mov   ebp, esp
  push  eax
  push  ebx
  push  edi
  mov   ebx, [ebp+8]        # ID-number for the gate
  call  get_idt_base
  lea   edi, [eax][ebx *8]  # gate's offset-address
  mov   eax, [ebp+0xc]
  mov   bx, [edi][0]        # entry-point's loword
  mov   [eax], bx
  mov   bx, [edi][6]        # entry-point's hiword
  mov   [eax][2], bx
  mov   eax, [ebp+0x10]
  mov   bx, [edi][2]        # code-segment selector
  mov   [eax], bx
  pop   edi
  pop   ebx
  pop   eax
  leave
  ret

#------------------------------------------------------------------------------
# Return the current CS selector in ax
FUNCTION(my_cs)
  mov  ax, cs
  ret

#------------------------------------------------------------------------------
FUNCTION(gpfHandler)
  push  ebp
  mov   ebp, esp
  push  eax

  # Assumption: DS is already using a selector with 4GB limit and 0 base
  # If not then preserve DS and setup a valid data descriptor

  # Stack at this point
  # [ebp]     Saved value of EBP register (from push EBP above)
  # [ebp+04h] Error code from gpf
  # [ebp+08h] Offset of executing instruction that caused gpf
  # [ebp+0Ch] Segment of executing instruction that caused gpf
  # [ebp+10h] Flags when executing instruction that caused gpf

  # verify gpf occurred at expected location
  mov   eax, [ebp+8]
  cmpd  [eax-8], 0x58475046  #'DWORD 'XGPF'
  jne   gpfHandler_exit

  # fix return address on stack to point at recovery code
  mov   eax, dword ptr [eax-4]
  mov   [ebp+8], eax

  gpfHandler_exit:
  pop   eax
  leave
  add   esp, 4      # Remove the error code from the stack
  iretd

FUNCTION(intHandler_asm)
  push  ebp
  mov   ebp, esp
  pushad

  # Assumption: DS is already using a selector with 4GB limit and 0 base
  # If not then preserve DS and setup a valid data descriptor

  # Stack at this point
  # [ebp]     Saved value of EBP register (from push EBP above)
  # [ebp+04h] Offset of executing instruction when interrupted
  # [ebp+08h] Segment of executing instruction when interrupted
  # [ebp+0Ch] Flags when interrupted

  call EXT_C(intHandler)

  popad
  leave
  iretd

FUNCTION(disable_interrupts)
  cli
  ret

FUNCTION(enable_interrupts)
  sti
  ret

#------------------------------------------------------------------------------
FUNCTION(cpuid_all)
# prototype -- void cpuid_all(U32 function, DWORD_REGS * regs)
# parameters -- function:DWORD, regs:PTR DWORD

  push  ebp
  mov   ebp, esp
  push  esi
  push  eax
  push  ebx
  push  ecx
  push  edx
  push  esi
  mov   eax, [ebp+8]    # CPUID function number
  mov   esi, [ebp+0xc]  # address of regs structure
  cpuid
  mov   [esi], eax
  mov   [esi][0x4], ebx
  mov   [esi][0x8], ecx
  mov   [esi][0xC], edx
  pop   esi
  pop   edx
  pop   ecx
  pop   ebx
  pop   eax
  leave
  ret

#------------------------------------------------------------------------------
FUNCTION(switch_stack_and_call)
  push  ebp
  mov   ebp, esp

  mov   ebx, [ebp+0xc]  # param
  mov   ecx, [ebp+0x8]  # mp_worker
  mov   esp, [ebp+0x10] # stack_top
  push  ebx
  call  ecx

  # never returns here

#------------------------------------------------------------------------------
FUNCTION(do_nothing)
  pause:
  jmp do_nothing


#------------------------------------------------------------------------------
FUNCTION(drop_ap_lock)
  push  ebp
  mov   ebp, esp

  push  eax
  push  ebx
  mov   ebx, dword ptr [ebp+0x08]
  lock  inc dword ptr [ebx][ASLEEP]
  xor   eax,eax
  xchg  dword ptr [ebx][BLOCK], eax
  pop   ebx
  pop   eax
  leave
  ret


#------------------------------------------------------------------------------
FUNCTION(pause32)
  pause
  ret

#------------------------------------------------------------------------------
FUNCTION(rmcb_start)
    cli
1:
    hlt
    jmp 1b
VARIABLE(rmcb_size)
    .long .-rmcb_start

#------------------------------------------------------------------------------
FUNCTION(ApStart)
  # On entry, ebx has the physical address of the SIPI block
  lock  inc dword ptr [ebx][AWAKE]

  mov   eax,1
TestLock:
  cmp   dword ptr [ebx][BLOCK], 1
  jne   TryGetLock
  pause
  jmp   TestLock

TryGetLock:
  xchg  dword ptr [ebx][BLOCK], eax
  test  eax, eax
  je    LockObtained
  pause
  jmp   TestLock

LockObtained:
  mov   ecx, ebx
  add   ecx, AP_STACK
  mov   esp, ecx

  # Save base pointer used to reference MP init shared data block
  push  ebx

  push  dword ptr [ebx][PARAM]
  call  dword ptr [ebx][FUNCTIONPTR]
  add   esp, 4

  # Restore base pointer used to reference MP init shared data block
  pop   ebx

  lock inc dword ptr [ebx][ASLEEP]

unblocking:
  xor   eax,eax
  xchg  dword ptr [ebx][BLOCK], eax

waiting:
  mov   eax, dword ptr [ebx][BARRIER]
  test  eax, eax
  jne   waiting

   #
   # processors enter Wait-for-SIPI state
   # Sends INIT IPI to itself to force Wait to SIPI state
   #
  mov   ecx, 0x01b             # Read APICBASE MSR
  rdmsr
  mov   edx, eax
  and   edx, 0x0400           # Test Extended XAPIC mode
  jnz   x2enabled
  and   eax, 0x0FFFFF000
  mov   esi, eax             # ESI -> Local APIC Base
  mov   eax, [esi+0x20]      # Get local APIC ID
  mov   [esi+0x310], eax     # Set physical destination for INIT IPI
  mov   eax, (0 SHL 18) OR (1 SHL 14) OR (5 shl 8)  # No destination shortcut OR LEVEL is 1 OR INIT IPI
  mov   [esi+0x300], eax     # Put this thread to sleep
forever:
  cli
  hlt                        # Halt.
  jmp forever                # Do this forever.
x2enabled:
  mov   ecx, 0x802
  rdmsr                     # eax has x2apic ID
  mov   edx, eax
  mov   eax, (0 SHL 18) OR (1 SHL 14) OR (5 shl 8) # No destination shortcut OR LEVEL is 1 OR INIT IPI
  mov   ecx, 0x830
  wrmsr
  jmp forever

# void wait_for_control_asm(U32 *control, U32 value, U32 use_mwait, U32 mwait_hint, U32 int_break_event)
FUNCTION(wait_for_control_asm)
  push ebp
  mov ebp, esp
  pushad

  mov esi, [ebp+8]
  mov ebx, [ebp+0xc]
  mov edi, [ebp+0x14]

  cmp dword ptr [ebp+0x10], 0
  je no_mwait

  cmp dword ptr [ebp+0x18], 1
  je mwait_int_break_event

mwait_no_int_break_event:
  cmp [esi], ebx
  je done
  mov eax, esi
  xor ecx, ecx
  xor edx, edx
  monitor
  mov eax, edi
  cmp [esi], ebx
  je done
  mwait
  jmp mwait_no_int_break_event

mwait_int_break_event:
  cmp [esi], ebx
  je done
  mov eax, esi
  xor ecx, ecx
  xor edx, edx
  monitor
  mov eax, edi
  mov ecx, 1
  cmp [esi], ebx
  je done
  mwait
  jmp mwait_int_break_event

no_mwait:
  cmp [esi], ebx
  je done
  pause
  jmp no_mwait

done:
  popad
  leave
  ret

VARIABLE(wait_for_control_asm_size)
    .long .-wait_for_control_asm    # Size, in bytes

.code16
#------------------------------------------------------------------------------
FUNCTION(pm32)
  #
  # Switch to Protected mode.
  #
  xor     ebx, ebx
  mov     bx, cs
  mov     ds, bx
  shl     ebx, 4   # Adjust segment into physical address format

  mov     eax,ebx
  .equ REL_SIPI_CODE_DESCRIPTOR, offset SIPI_CODE_DESCRIPTOR - offset pm32
  mov     si, REL_SIPI_CODE_DESCRIPTOR
  mov     [si+0x2], ax                       # Put lower physical address into GDT.
  shr     eax,0x10                           # High nibble of physical address only.
  mov     [si+0x4], al                       # Put high nibble physical address into GDT.

  mov     eax,ebx
  .equ REL_BootGDTtable, offset BootGDTtable - offset pm32
  add     eax, REL_BootGDTtable          # Add in offset of GDT table.
  .equ REL_GDTDescriptor, offset GDTDescriptor - offset pm32
  mov     si, REL_GDTDescriptor
  mov     [si+0x2], eax                       # Store in GDT address variable.

  .byte   0x66
  lgdt    cs:REL_GDTDescriptor

  mov     eax, cr0                         # Get control register 0
  or      eax, 0x00000003                  # Set PE bit (bit #0) & MP bit (bit #1)
  mov     cr0, eax                         # Activate protected mode
  mov     eax, cr4                         # Get control register 4
  or      eax, 0x00000600                  # Set OSFXSR bit (bit #9) & OSXMMEXCPT bit (bit #10)
  mov     cr4, eax

  # Immediately following the MOV CR0 instruction, execute a far JMP or far
  # CALL instruction. This operation is typically a far jump or call to the
  # next instruction in the instruction stream.
  .byte  0x0EA
.equ REL_jmp_target, 1f - offset pm32
  .word  REL_jmp_target
  .word  SIPI_CODE_SEL
1:

  # Set up selectors for Protected Mode entry.
.att_syntax
  mov     $SYS_DATA_SEL, %ax
.intel_syntax noprefix
  mov     ds, ax
  mov     es, ax
  mov     fs, ax
  mov     gs, ax
  mov     ss, ax

  #
  # Go to Protected32
  #
.equ REL_CpuRefCodeLinearAddress, offset CpuRefCodeLinearAddress - offset pm32
  jmp     FWORD PTR cs:[REL_CpuRefCodeLinearAddress]

.align 4

# 16-bit Selector and 32-bit offset value for the FAR JMP to Processor Reference Code.
VARIABLE(CpuRefCodeLinearAddress)
  .long  offset ApStart
  .word  LINEAR_CODE_SEL

.align 16

# GDT[0]: 0x00: Null entry, never used.
.equ NULL_SEL, .-GDT_BASE            # Selector [0]
GDT_BASE:
VARIABLE(BootGDTtable)
  .long 0
  .long 0

# Linear code segment descriptor
.equ LINEAR_CODE_SEL, .-GDT_BASE    # Selector [0x8]
  .word 0x0FFFF # limit 0xFFFFF
  .word 0       # base 0
  .byte 0
  .byte 0x09B   # present, ring 0, code, execute/read, non-conforming, accessed
  .byte 0x0CF   # page-granular, 32-bit
  .byte 0

# System data segment descriptor
.equ  SYS_DATA_SEL, .-GDT_BASE    # Selector [0x10]
  .word 0x0FFFF # limit 0xFFFFF
  .word 0       # base 0
  .byte 0
  .byte 0x093   # present, ring 0, data, read/write, expand-up, accessed
  .byte 0x0CF   # page-granular, 32-bit
  .byte 0

# SIPI code segment descriptor
.equ  SIPI_CODE_SEL, .-GDT_BASE # Selector [0x18]
SIPI_CODE_DESCRIPTOR:
  .word 0x0FFFF # limit 0xFFFF
  .word 0x0000  # base
  .byte 0x00
  .byte 0x09B   # present, ring 0, code, execute/read, non-conforming, accessed
  .byte 0x000   # byte-granular, 16-bit
  .byte 0

.equ  GDT_SIZE, .-BootGDTtable  # Size, in bytes

# Global Descriptor Table Descriptor
GDTDescriptor:      # GDT descriptor
  .word GDT_SIZE-1  # GDT limit
  .word 0x0000      # GDT base address
  .word 0x0000

VARIABLE(pm32_size)
    .long .-pm32   # Size, in bytes

